#
# Copyright (c) Daniel Sheffield 2021
#
# All rights reserved
#
# THIS SOFTWARE IS PROVIDED AS IS WITHOUT WARRANTY
from .. import COPYRIGHT
from ..widgets import (
    AutoCompletePopUp,
    NoTabCheckBox,
    AutoCompleteEdit,
    AutoCompleteFloatEdit,
    FocusWidget
)
from collections import (
    OrderedDict,
)

import urwid
class TransactionEditor(FocusWidget):

    def keypress(self, size, key):
        if isinstance(key, tuple):
            return
        
        if getattr(self.original_widget.original_widget, 'original_widget', None) is None:
            return super().keypress(size, key)
        
        if key == 'tab':
            self.advance_focus()
        elif key == 'shift tab':
            self.advance_focus(reverse=True)
        else:
            return super().keypress(size, key)

    def _init_data(self, fields):
        self.data = OrderedDict([
            (k, '') for k in fields
        ])
        self.clear()

    def _apply_choice(self, name, value):
        self._apply_changes(name, value)
        for k,v in self.data.items():
            if k == name or v:
                continue
            options = self.query_manager.unique_suggestions(k, **self.data)
            if len(options) == 1 and k != 'ts':
                self._apply_changes(k, list(options)[0])
    
    def apply_choice(self, name):
        return  lambda w,x: self._apply_choice(name, x)

    def _apply_changes(self, name, value):
        self.data.update({
            name: value,
        })
    
    def apply_changes(self, name):
        return lambda w,x: self._apply_changes(name, x)

    def apply_organic_state(self, w, state):
        self.data['organic'] = repr(state).lower()

    def clear(self):
        for k in self.data:
            if k in ('ts', 'store',):
                continue
            self.data[k] = ''
        self.organic_checkbox.set_state(False)
    
    def update(self):
        for k in self.edit_fields:
            self.edit_fields[k].set_edit_text(self.data[k])
        date, store = self.data['ts'], self.data['store']
        self.text_fields['dbview'].set_text(
            self.query_manager.get_session_transactions(date, store) if None not in (
                date or None, store or None
            ) else ''
        )
        self.organic_checkbox.set_state(True if self.data['organic'] == 'true' else False)
        return self

    def __init__(self, query_manager, fields,
        layout, side_pane, bottom_pane,
        save_and_clear_cb, autocomplete_cb, apply_choice_cb):
        self.organic_checkbox = NoTabCheckBox(
            u"Organic",
            on_state_change=self.apply_organic_state
        )
        self.query_manager = query_manager
        self._init_data(fields)
        self.layout = layout
        self.side_pane = side_pane
        self.bottom_pane = bottom_pane
        self.edit_fields = OrderedDict()
        self.text_fields = OrderedDict()
        for k in self.data:
            if k in self.side_pane and k != 'unit':
                ef = AutoCompleteFloatEdit(('bg', k))
            elif k != 'organic':
                ef = AutoCompleteEdit(('bg', k))
            else:
                continue

            ef.set_edit_text(self.data[k])
            urwid.connect_signal(ef, 'change', self.apply_changes(k))
            self.edit_fields[k] = ef
    
        header = urwid.Text(u'Fill Transaction', 'center')
        _copyright = urwid.Text(COPYRIGHT, 'center')
        done_button = urwid.Button(('streak', u'Done'))
        urwid.connect_signal(done_button, 'click', lambda w: save_and_clear_cb())
        banner = urwid.Pile([
            urwid.Padding(header, 'center', width=('relative', 100)),
            urwid.Padding(_copyright, 'center', width=('relative', 100)),
        ])
        banner = urwid.AttrMap(banner, 'banner')
        fields = dict([
            (k, urwid.LineBox(urwid.AttrMap(AutoCompletePopUp(
                    self.edit_fields[k],
                    apply_change_func=lambda name, pop_up_cb: autocomplete_cb(name, self.data, pop_up_cb),
                    apply_choice_cb=apply_choice_cb,
                ), 'streak'), title=k.title(), title_align='left')) for k in self.edit_fields if k != 'product'
        ])
        fields.update({
            'product': urwid.LineBox(urwid.Columns([
                        urwid.AttrMap(AutoCompletePopUp(
                            self.edit_fields['product'],
                            apply_change_func=lambda name, pop_up_cb: autocomplete_cb(name, self.data, pop_up_cb),
                            apply_choice_cb=apply_choice_cb,
                        ), 'streak'),
                self.organic_checkbox
            ], dividechars=2), title='Product', title_align='left')
        })
        txn_view = urwid.Text('')
        self.text_fields.update({'dbview': txn_view})
        fields.update({
            'dbview': urwid.LineBox(
                urwid.AttrMap(txn_view, 'streak'),
                title="Session Data",
                title_align='left',
            )
        })
        
        side_pane_widget = (12, urwid.Pile([
            fields[r] if r is not None else urwid.Divider() for r in self.side_pane
        ]))
        main_pane_widgets = []
        for i, r in enumerate(self.layout):
            widgets = []
            for c in r:
                if c is not None:
                    if c == 'organic':
                        continue
                    widgets.append(fields[c])
                else:
                    widgets.append(urwid.Divider())
            main_pane_widgets.append(urwid.Columns(widgets))


        main_pane_widget = urwid.Pile(main_pane_widgets)
        
        widget = urwid.Pile([
            banner,
            urwid.Divider(),
            urwid.Columns((main_pane_widget, side_pane_widget),
                dividechars=2,
            ),
            *[ fields[c] if c is not None else urwid.Divider() for c in self.bottom_pane ],
            urwid.Divider(),
            done_button,
        ])
        widget = urwid.Filler(widget, 'top') 
        widget = urwid.AttrMap(widget, 'bg')
        super().__init__(widget, [2,0,0,0,0], [
            [2, 0, 1, 0],
            [4,],
            [5,],
            [7,],
        ])
